; raw sector mfm loader
;
; TOKI modification:
; a0=file info (buffer.L,start sector*512.L,len sector*512.L)

; bits CIAA-PRA --------------------------- bits signification
DSKRDY	=	5		;drive ready
DSKTRK0	=	4		;track 0 reached
DSKPROT	=	3		;disk write protected
DSKCHNG	=	2		;disk change
; bits CIAB-PRB ---------------------------
DSKMTR	=	7		;control motor
DSKSEL3	=	6		;select DF3
DSKSEL2	=	5		;	DF2
DSKSEL1	=	4		;	DF1
DSKSEL0	=	3		;	DF0
DSKSIDE	=	2		;select side/head
DSKDIR	=	1		;direction
DSKSTEP	=	0		;move head

TRKLEN=11*512
SYNC=$4489

DOSTRACKLOADER:
		movem.l	d0-a6,-(sp)
		lea	$DFF000,a6	; CUSTOM
		lea	$BFD100,a4	; CIAB-PRB / +$F01 = CIAA-PRA
		; init CIA Timer B
		move.b	#$00,$500(a4)	; timer B lo $bfd600
		move.b	#$0d,$600(a4)	;         hi $bfd700
		clr.b	$e00(a4)	; crb

		lea	track(pc),a2	; current track position
		lea	$1000.w,a1	; MFM buffer
		; TOKI file infos
		move.l	4(a0),d0	; sector offset
		move.l	8(a0),d1	; len bytes
		movea.l	(a0),a0		; dest buffer

		movea.l	a0,a5
		adda.l	d1,a5		; dest buffer+len

		move.b	#$7D,(a4)	; start drive %01111101
		bclr	#DSKSEL0,(a4)	; DSKSEL0 select DF0
		bsr	Delay

		clr.l	d4
		divu.w	#TRKLEN,d0	; (start sector*$1600)/$1600 = num sector|num track (ex. 13*$1600) 
		swap	d0		; (ex. d0=$0400 0001 => $0001 0400)
		move.w	d0,d4		; start sector in track 0-10 (rest of division= $400 = 2*512)
		swap	d0		; num track (ex. $0400 0001 d0=1)
		bsr	SeekTrack	; move head to track

		moveq	#6-1,d5		; try
.load		move.l	a1,$20(a6)	; DSKPTH MFM Buffer
		move.w	#$8210,$96(a6)	; DMACON disk on
		move.w	#$7F00,$9E(a6)	; erase ADKCON
		move.w	#$9500,$9E(a6)	; correct value of ADKCON
		move.w	#$4000,$24(a6)	; erase DSKLEN
		move.w	#SYNC,$7E(a6)	; DSKSYNC standard Sync	Word

		bsr.s	DiskReady

		move.w	#$8000+$1900,d1	; track	size standard 9900
		move.w	d1,$24(a6)	; DSKLEN
		move.w	d1,$24(a6)	; 2 times to start DMA transfert
		move.w	#2,$9C(a6)	; INTREQ dsk interrupt

.wait		btst	#1,$1F(a6)	; INTREQR disk ready ?
		beq.s	.wait		; not yet

		move.w	#0,$24(a6)	; erase DSKLEN
		
		bsr.w	MFMUncode	; decode mfm buffer
		bne.s	.error
		cmpa.l	a0,a5		; end ?
		ble.s	.finished	; yep

		bsr.w	NextTrack	; next
		bra.s	.load

.finished	move.b	#$FD,(a4)	; stop drive %11111101
		bclr	#DSKSEL0,(a4)	; select DF0
		bset	#DSKSEL0,(a4)	; deselect DF0
		bsr	Delay
		movem.l	(sp)+,d0-a6
		rts

.error		subq.l	#1,d5		; try again ?
		bne.s	.load
;- error
infiniteloop:	move.w	d0,$180(a6)	; flash color0
		addq.l	#1,d0
		bra.s	infiniteloop
;------------------------------------------- 
; Functions
;-------------------------------------------
DiskReady:	btst	#DSKRDY,$F01(a4)	; CIAA-PRA
		bne.s	DiskReady
		rts
;--------
Seek0:		btst	#DSKTRK0,$F01(a4)	; CIAA-PRA DSKTRK0
		beq.s	.found
		bset	#DSKDIR,(a4)	; DSKDIR goto 0
		bsr.s	MoveHead
		bra.s	Seek0
.found		clr.b	(a2)		; track 0
		rts
;------
NextTrack:	addq.b	#1,(a2)		; current track+1
		move.b	(a2),d1
		btst	#0,d1		; odd or even track ?
		bne.s	.changeSide
		bclr	#DSKDIR,(a4)	; DSKDIR goto 79
		bsr.s	MoveHead
.changeSide	bchg	#DSKSIDE,(a4)	; DSKSIDE switch side 0/1
		bsr.w	Delay
		rts	
;------
MoveHead:	bset	#DSKSTEP,(a4)	; inactive
		nop	
		nop	
		bclr	#DSKSTEP,(a4)	; active
		nop	
		nop	
		bset	#DSKSTEP,(a4)	; inactive
		bsr.w	Delay		; max 3 ms
		bra.s	DiskReady
;-------------
SeekTrack:	movem.l	d1-d2,-(sp)

		cmpi.b	#-1,(a2)	; current track ok ?
		bne.s	.ok
		bsr.s	Seek0		; no. seek track 0
.ok		clr.l	d1
		clr.l	d2
		move.b	(a2),d1		; current track
		move.b	d0,d2		; wanted track
		bset	#DSKSIDE,(a4)	; DSKSIDE side 0
		btst	#0,d0		; odd or even track
		beq.s	.side0
		bclr	#DSKSIDE,(a4)	; DSKSIDE side 1

.side0		lsr.w	#1,d2		; track/2 (80 tracks/side)
		lsr.w	#1,d1		; current track/2 
		sub.w	d1,d2		; track to reach-current track = direction
		bcs.s	.negat		; <0
		bhi.s	.posit		; >0
		bra.s	.found		; =0

.posit		bclr	#DSKDIR,(a4)	; DSKDIR goto 79
		bra.s	.moveh

.negat		neg.w	d2		; positive
		bset	#DSKDIR,(a4)	; DSKDIR goto 0

.moveh		bsr.s	MoveHead
		subq.w	#1,d2		; goto the track
		bne.s	.moveh

.found		move.b	d0,(a2)
		movem.l	(sp)+,d1-d2
		rts	
;----------------------------------------------
; standard MFM format (see explanation below) 
; a0=dest buffer
; a1=mfm buffer
MFMUncode:	movem.l	a1-a6,-(sp)
		movea.l	a0,a2		; dest buffer
		movea.l	a0,a3
		adda.l	d4,a3		; dest buffer+start sector
		movea.l	a0,a4
		cmpi.w	#SYNC,(a1)	; find sync word
		bne.s	.error

		moveq	#11-1,d5	; 11 sectors
.loopsector:	bsr.s	UncodeHeader
		moveq	#128-1,d6	; 128 words to decode (512/4)
.UncodeSector	move.l	512(a1),d1	; even bits
		move.l	(a1)+,d0	; odd bits
		eor.l	d0,d2
		eor.l	d1,d2
		bsr.s	uncode
		cmpa.l	a3,a2		; real start reached ?
		bge.s	.ok
		addq.l	#4,a2		; not yet
		bra.s	.next		; skip sector

.ok		movea.l	a2,a6
		addq.l	#4,a2
		suba.l	d4,a6
		cmpa.l	a5,a6		; end of dest buffer reached ?
		bge.s	.next		; yes. skip sector
		move.l	d0,(a6)+	; store decoded data

.next		dbf	d6,.UncodeSector

		andi.l	#$55555555,d2	; mfm bits Mask %010101...
		cmp.l	d2,d7
		bne.s	.error
		dbf	d5,.loopsector

		; sector precision : 
		; track 1 [**|*********] start at sector 13  
		; len=56320. start=(13*512)/(11*512) = 13/11 = 1.2 = track 1, 2 sector
		; 5632-2*512. start buffer+2*512
		move.l	#TRKLEN,d0	; standard size 11*512
		sub.l	d4,d0		; - start sector
		adda.l	d0,a0		; skip lower sectors
		clr.l	d0
		clr.l	d4

.exit:		movem.l	(sp)+,a1-a6
		tst.l	d0
		rts	

.error:		movea.l	a4,a0
		moveq	#-1,d0
		bra.s	.exit
;--------------------
; a0=dest data
; a1=source mfm data
UncodeHeader:	move.w	#SYNC,d1

.GAP:		cmp.w	(a1)+,d1	; search start of sector
		bne.s	.GAP
		cmp.w	(a1),d1		; another sync word ?
		bne.s	.ok
		addq.l	#2,a1		; skip second sync word
.ok:		move.l	48(a1),d0	; Datas checksum (odd bits)
		move.l	52(a1),d1	; Datas checksum (even bits)
		bsr.w	uncode
		move.l	d0,d7
		move.l	(a1),d0		; ID sector (odd bits)
		move.l	4(a1),d1	; ID sector (even bits)
		bsr.s	uncode		; d0=format,track,sector,count
		lsr.w	#8,d0		; shift 8 bits to right to get sector
		lea	56(a1),a1	; point to Datas area
		add.w	d0,d0		; *2
		asl.w	#8,d0		; *256
		lea	(a4,d0.w),a2	; point to sector start
		clr.l	d2
		rts
;----------------------
uncode:		asl.l	#1,d0
		andi.l	#$AAAAAAAA,d0	; even bits Mask
		andi.l	#$55555555,d1	; odd bits Mask
		or.l	d1,d0		; merge odd and even bits
		rts	
;------------------------------------------
; CIA TIMER DELAY
Delay		move.b	#$19,$e00(a4)	;$bfdf00 start timer
		nop
		nop
.dly:		btst.b	#0,$e00(a4)	; and wait
		bne.s	.dly
		rts
; var		
track		dc.b	-1,0	; current track
		
;----------------------------------------- MFM Format informations
; principle: 
; insert a parity bit between each real data bit.
; why? 
; It helps avoid any synchronization error while I/O.
; The odd bits are coded first followed by even bits.
;
; Datas	MFM
; 1	01
; 0	10 (if preceded by 0)
; 0	00 (if preceded by 1)
;
; A word synchronization is placed in front of each sector in order to recognize the start.
; It corresponds to no possible coding to know that it's a mark of sync and not a data.
; The standard one is $4489, but others words can be used.
; 
; Track format:
; 11 sectors with a data area infos of 32 bytes followed by a data area of 512 bytes.
; Before the track, there is a GAP of variable size (approx. 700 bytes).
; This GAP avoid to crush last datas with first (tracks are circulare)
;
; offset	value	signification
;   0		$AAAA	first sector word
;   2		$AAAA	second sector word
;   4		$4489	first sync word
;   6		$4489	second sync word
;   8		xxxxx	ID sector (odd bits)
;  12		xxxxx	ID sector (even bits)
;  16		xxxxx	Dos datas (odd bits)
;  32		xxxxx	Dos datas (even bits)
;  48		xxxxx	Header checksum (odd bits)
;  52		xxxxx	Header checksum (even bits)
;  56		xxxxx	Datas checksum (odd bits)
;  60		xxxxx	Datas checksum (even bits)
;  64		xxxxx	Datas (odd bits)
; 576		xxxxx	Datas (even bits)
